"""
Default instances for dynamic runs.

Typically, developers will only run one digital worker on a Python
runtime, and it is tricky to initialize an object and pass that object
instance down into other files in a project for a custom decorator.
This package provides a set of default instances of objects and contexts
for dynamic runs.

This file initializes a shared `ReportBuilder` object, which is used
individual steps and sub steps to share state.

See the quickstart for, well, quickstarts on each of these methods.
"""

import logging
import pathlib
import warnings
from typing import Union

from thoughtful.environment_variables import EnvironmentVariables
from thoughtful.supervisor.event_bus import EventBus
from thoughtful.supervisor.main_context import MainContext
from thoughtful.supervisor.manifest import Manifest
from thoughtful.supervisor.reporting.report_builder import ReportBuilder
from thoughtful.supervisor.step_context import StepContext
from thoughtful.supervisor.step_decorator_factory import create_step_decorator
from thoughtful.supervisor.streaming.jwt_auth import JWTAuth
from thoughtful.supervisor.streaming.streamer import Streamer
from thoughtful.supervisor.streaming.token import Token

logger = logging.getLogger(__name__)

#: A shared ``EventBus`` for any ``Context`` or step decorators to publish
#: events to.
shared_bus = EventBus()


def connect_to_fabric(
    webhook_url: str,
    access_token: str,
    refresh_token: str,
    refresh_url: str,
) -> None:
    """
    Connect this worker to an instance of Fabric and stream updates to the
    ``webhook_url`` endpoint. (Fabric is Empower's backend API.)

    Each of these values should be available in your Robocorp environment
    via the work item's ``variables`` object.

    Args:
        webhook_url (str): The URL to stream events to. This is a Fabric
            endpoint, and it will likely be in the form of
            ``https://fabric.thoughtful.ai/webhooks/users-processes-updates-jwt``.
        access_token (str): The encoded JWT access token for the Fabric user.
        refresh_token (str): The encoded JWT refresh token for the Fabric user.
        refresh_url (str): The URL to refresh the Fabric user's JWT tokens. This
            is a Fabric endpoint, and it will likely be in the form of
            ``https://fabric.thoughtful.ai/refresh-token``.
    """
    run_id = EnvironmentVariables().run_id
    if not run_id:
        warnings.warn(
            "No run ID found in environment variables, so I can't stream" "to Fabric."
        )
        return

    fabric_auth = JWTAuth(
        access_token=Token(encoded_value=access_token),
        refresh_token=Token(encoded_value=refresh_token),
        refresh_url=refresh_url,
    )
    fabric_streamer = Streamer(
        run_id=EnvironmentVariables().run_id,
        callback_url=webhook_url,
        auth=fabric_auth,
    )
    shared_bus.subscribe(fabric_streamer.handle_event)


#: A shared ``ReportBuilder`` for steps and step contexts to add step reports
#: to.
report_builder = ReportBuilder()

#: Use this decorator on your own functions to mark what workflow step each
#: Python function is matched to.
step = create_step_decorator(report_builder=report_builder, event_bus=shared_bus)

#: Use this function to mark a step as failed.
#: This is just a shortcut for ``set_step_status(step_id, "failed")``.
#: Exposes ``report_builder.fail_step``.
fail_step = report_builder.fail_step

#: Expose the report builder's ability to override a step's status as a
#: top-level call. Exposes ``report_builder.set_step_status``.
set_step_status = report_builder.set_step_status

#: Expose the report builder's ability to override a step's record's status as a
#: top-level call. Exposes ``report_builder.set_record_status``.
set_record_status = report_builder.set_record_status

#: Expose the report builder's ability to override a run's status as a
#: top-level call. Exposes ``report_builder.set_run_status``.
set_run_status = report_builder.set_run_status


# noinspection PyPep8Naming
class step_scope(StepContext):
    """
    It's a context manager that provides a scope for a step using the
    aforementioned default instances of ``report_builder`` and ``recorder``.
    """

    def __init__(self, step_id: str):
        """
        A default `StepContext` that uses the root level `report_builder` and
        `recorder`.

        Args:
            *step_id: The list of integers that represent the step ID.
        """
        super().__init__(
            builder=report_builder,
            step_id=step_id,
            event_bus=shared_bus,
        )


# noinspection PyPep8Naming
class supervise(MainContext):
    """
    It's a context manager that provides a scope for the main context using the
    aforementioned default instances of ``report_builder`` and ``recorder``.
    """

    def __init__(
        self,
        manifest: Union[Manifest, str, pathlib.Path] = "manifest.yaml",
        output_dir: Union[str, pathlib.Path] = "output/",
        *args,
        **kwargs,
    ):
        """
        A default `MainContext` that uses the root level `report_builder` and
        `recorder`.

        Args:
            *args: Extra arguments to the `MainContext` constructor.
            **kwargs: Extra keyword arguments to the `MainContext` constructor.
            manifest (str): The digital worker's manifest definition
            output_dir (str): Where the work report and drift report will
                be written to
        """
        super().__init__(
            report_builder=report_builder,
            manifest=manifest,
            output_dir=output_dir,
            upload_uri=EnvironmentVariables().s3_bucket_uri,
            event_bus=shared_bus,
            *args,
            **kwargs,
        )


if __name__ == "__main__":
    from thoughtful.supervisor.event_bus import RunStatusChangeEvent
    from thoughtful.supervisor.reporting.status import Status

    _streamer = Streamer.from_encoded_tokens(
        run_id=EnvironmentVariables().run_id,
        callback_url=EnvironmentVariables().callback_url,
        # The JWT token is set by the user later. If steamer is set to None,
        # then the step decorators can't be updated to stream if streamer is
        # changed to a new instance.
        #
        # See CX-2368 for the proper fix by lifting the streamer out of the contexts
        access_token="",
        refresh_token="",
        refresh_url="",
    )
    shared_bus.subscribe(_streamer.handle_event)
    test = RunStatusChangeEvent(status=Status.RUNNING)
    shared_bus.emit(test)



"""
The step context is a context manager that is used to supervise a single step
of a digital worker. It is used in lieu of the ``@step`` decorator for
situations when you do not want to write a function for a step. For example,
it is very useful when iterating over a list of items:

.. code-block:: python

    from thoughtful.supervisor import supervise, step_scope

    def times_two(integers: list) -> None:
        return integers * 2

    def times_three(integer: int) -> None:
        return integer * 3

    def main():
        num_list = [1, 2, 3, 4, 5]

        with step_scope("1.1"):
            num_list = times_two(num_list)

        for num in num_list:
            with step_scope("1.2"):
                num = times_three(num)

    if __name__ == '__main__':
        with supervise():
            main()

Aside: There are a few scenarios where this is actually more convenient than
using the decorator on a lower level. Thanks to the fact that we can yield
the ``self`` object from a context manager, we can use this to update
attributes such as the record status, the step status, and is even a bit
cleaner when setting the record ID.
"""

from __future__ import annotations

from types import TracebackType
from typing import Optional, Type, Union

from thoughtful.supervisor.event_bus import EventBus, StepReportChangeEvent
from thoughtful.supervisor.reporting.report_builder import ReportBuilder
from thoughtful.supervisor.reporting.report_builder import StepReportBuilder
from thoughtful.supervisor.reporting.status import Status
from thoughtful.supervisor.reporting.timer import Timer


class StepContext:
    """
    A context manager for a step that is running inside another step.
    This is an alternative to ``@step`` decorator when you don't want
    to write an entire function for a step.
    """

    def __init__(
        self,
        builder: ReportBuilder,
        step_id: Union[str, int],
        event_bus: EventBus,
    ):
        """
        Args:
            builder (ReportBuilder): Where the step report will be written.
            step_id (str): The step id of this step, ie `"1.1"`
            event_bus (EventBus): The decorator will emit events, such as when
                a step report has been created or updated, to this bus.
        """
        self.step_id = str(step_id)
        self.report_builder = builder
        self.timer = Timer()
        self._status_override: Optional[Status] = None
        self.event_bus = event_bus
        self.step_report_builder: StepReportBuilder

    def __enter__(self):
        """
        Logic for when this context is first started.

        Returns:
            MainContext: This instance.
        """
        start_time = self.timer.start()
        self.step_report_builder = StepReportBuilder(
            step_id=self.step_id,
            start_time=start_time,
            status=Status.RUNNING,
        )

        for report in self.step_report_builder.to_reports():
            new_event = StepReportChangeEvent(step_report=report)
            self.event_bus.emit(new_event)

        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        """
        Runs when the context is about to close, whether caused
        by a raised Exception or now.

        Returns:
            bool: True if the parent caller should ignore the
                Exception raised before entering this function
                (if any), False otherwise.
        """
        timed_info = self.timer.end()
        if self._status_override:
            step_status = self._status_override
        else:
            step_status = Status.FAILED if exc_type else Status.SUCCEEDED

        # Update step report with finished step details
        self.step_report_builder.end_time = timed_info.end
        self.step_report_builder.status = step_status

        for report in self.step_report_builder.to_reports():
            new_event = StepReportChangeEvent(step_report=report)
            self.event_bus.emit(new_event)

        self.report_builder.step_report_builders.append(self.step_report_builder)

        # Return False so that any exceptions inside this context
        # are still raised after this function ends
        return False

    def error(self) -> None:
        """
        Sets the status of this step to `Status.FAILED` in its `StepReport`.

        .. code-block:: python

            with step_scope("1.1") as s:
                ...  # do some stuff
                s.error()

        .. code-block:: json

            {
                "workflow": [
                    {
                        "step_id": "1.1",
                        "step_status": "failed"
                    }
                ]
            }
        """
        self.set_status(Status.FAILED)

    def set_status(self, status: Union[str, Status]) -> None:
        """
        Override the step context's status to be in the status of ``status``

        Args:
            status (str, Status): The status to set the step to

        .. code-block:: python

            with step_scope("1.1") as s:
                ...  # do some stuff
                s.set_status("warning")

        .. code-block:: json

            {
                "workflow": [
                    {
                        "step_id": "1.1",
                        "step_status": "warning"
                    }
                ]
            }
        """
        # Convert the status to the correct type if necessary
        safe_status = Status(status)
        self._status_override = safe_status

    def set_record_status(
        self,
        status: Union[str, Status],
        record_id: str = "",
        message: Optional[str] = None,
        metadata: Optional[dict] = None,
    ) -> None:
        """
        Override a step context's record to be in the status of ``status``

        Args:
            status (str, Status): The status to set the record to
            record_id (str): The ID of the record
            message (str, optional): A short message to attach to the record
            metadata (dict, optional): A dictionary of metadata to attach to the
                record. These should be short, tag-like values.

        .. code-block:: python

            with step_scope("1.1") as s:
                ...  # do some stuff
                s.set_record_status("warning", "record01")

        .. code-block:: json

            {
                "workflow": [
                    {
                        "step_id": "1.1",
                        "step_status": "succeeded",
                        "record": {
                            "id": "kaleb_cool_guy",
                            "status": "warning"
                        }
                    }
                ]
            }

        """
        # Convert the status to the correct type if necessary
        if message is not None and not type(message) == str:
            raise ValueError("Record message must be a string")
        if metadata is not None and not type(metadata) == dict:
            raise ValueError("Record metadata must be a dict")

        message = message or ""
        metadata = metadata or {}
        safe_status = Status(status)
        self.step_report_builder.set_record_status(
            record_id, safe_status, message, metadata
        )


if __name__ == "__main__":
    report_builder = ReportBuilder()

    substep = StepContext

    with substep(report_builder, 1) as s:
        print("hello world")

        with substep(report_builder, "1.1") as s2:
            print("inner step")

    print(report_builder.step_report_builders)



"""
This module holds the logic for the runtime storage and generation of the
run report. It contains metadata about the two main entities in the run report:
the run as a whole and its children, the steps.
"""

from __future__ import annotations

import json
import pathlib
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union

from thoughtful.__version__ import __version__
from thoughtful.supervisor.reporting.record_report import RecordReport
from thoughtful.supervisor.reporting.status import Status
from thoughtful.supervisor.reporting.step_report import StepReport
from thoughtful.supervisor.reporting.timed_report import TimedReport


@dataclass
class Report(TimedReport):
    """
    A report is a representation of a run's execution, and it includes metadata
    about the run, a list of the steps that were executed, and records that
    were processed in each step (if any). There will only be one report per run.
    """

    supervisor_version = str(__version__)
    """
    str: The version of the supervisor that generated the report.
    """

    workflow: List[Union[StepReport, RecordReport]] = field(default_factory=list)
    """
    List[StepReport]: The list of steps that were executed.
    """

    status: Optional[Status] = None
    """
    Status, optional: The status of the run.
    """

    def __json__(self) -> Dict[str, Any]:
        return {
            **super().__json__(),
            "supervisor_version": self.supervisor_version,
            "workflow": [step.__json__() for step in self.workflow],
            "status": self.status.value,
        }

    def write(self, filename: Union[str, pathlib.Path]) -> None:
        """
        Write the report as a JSON object to a file.

        Args:
            filename: Where to write the file.
        """
        path = (
            filename if isinstance(filename, pathlib.Path) else pathlib.Path(filename)
        )

        with path.open("w") as out:
            report_dict = self.__json__()
            json.dump(report_dict, out)



from __future__ import annotations

import logging
from dataclasses import dataclass

import requests
from requests.auth import AuthBase

from thoughtful.supervisor.streaming.token import Token

logger = logging.getLogger(__name__)

POST_TIMEOUT_SECONDS = 10


@dataclass
class JWTAuth(AuthBase):
    access_token: Token
    refresh_token: Token
    refresh_url: str

    def __call__(self, r: requests.Request) -> requests.Request:
        self.refresh()
        r.headers["Authorization"] = f"Bearer {self.access_token}"
        return r

    def refresh(self) -> None:
        if not self.access_token.is_expired:
            return

        logger.info("Access token expired, refreshing")
        response = requests.post(
            self.refresh_url,
            json={
                "refreshToken": str(self.refresh_token),
            },
            timeout=POST_TIMEOUT_SECONDS,
        )

        if not response.ok:
            logging.warning("Could not refresh JWT token!")
            logging.warning(
                f"Received response {response.status_code}: {response.text}"
            )
            return

        logging.info("Successfully refreshed JWT token")
        new_values = response.json()
        self.access_token = Token(new_values["accessToken"])
        self.refresh_token = Token(new_values["refreshToken"])



"""
Package collection for Thoughtful, built and maintained by Thoughtful.
"""
import warnings

from thoughtful.__version__ import __version__

__title__ = "thoughtful"

warnings.simplefilter("always", DeprecationWarning)



"""
This module contains the enum that is intended to be used to set the status of
anything, and everything in Supervisor. There may be edge cases, but these
three statuses should be enough to cover all the bases.
"""


from enum import Enum


class Status(str, Enum):
    """
    Status — The status of whatever is being tracked — step, entire flow, record
    """

    SUCCEEDED = "succeeded"
    """
    The item has completed successfully.
    """

    FAILED = "failed"
    """
    The item has failed.
    """

    WARNING = "warning"
    """
    The item requires attention.
    """

    RUNNING = "running"
    """
    The item is running.
    """
